#
#  evaluate.py contains the evaluator implementations for the meta-model derived from MATCH.
#  The meta-model element should have a function that returns the path to the implementation.
#  An Evaluator should be derived from MatchEvaluator and implement the same interface.
#  Look at the documentation string to find out what the functions do.
#

import sys
import coral
import coral.util.search as S
import coral.util.search.util as SU

from code import InteractiveInterpreter
import string
L = coral.lang.MATCH10

class II(InteractiveInterpreter):

    def __init__(self, locals = {}):
        self.fail = False
        self.message = []
        InteractiveInterpreter.__init__(self, locals)
        
    def write(self, data):
        self.message.append(data)
        self.fail = True
        
    def runcode(self, code):
        InteractiveInterpreter.runcode(self, code)

class MatchEvaluator:
    def __init__(self, matchElement):
        self.elem = matchElement

    def search(self, target):
        """Search elements in target that matches the Match-element """
        raise NotImplemented("MatchEvaluator.search should be overridden")
    
    def appliesTo(self, element):
        """Checks if this matches element"""
        raise NotImplemented("MatchEvaluator.appliesTo should be overridden")

    def matchesElement(self, match):
        """Runs the matchscript and returns a context"""
        raise NotImplemented("MatchEvaluator.matchesElement should be overridden")
        
    def __str__(self):
        """Returns the Match element as a string"""
        return str(self.elem)
    


class GraphMatchEvaluator(MatchEvaluator):
    
    def __init__(self, GMatch):
        MatchEvaluator.__init__(self, GMatch)
        self.match = GMatch
        self.query = GMatch.query
        self.types = coral.Set()
        for qelem in self.query.queryelement:
            if qelem.isActive:
                for part in qelem.abstractElement.__getAllParts__():
                    self.types.append(part.__type__())

        scriptList = []
        scriptList.append("import coral\ndef script():\n")
        for qelement in self.query.queryelement:
            scriptList.append("\taddToContext('"+qelement.name+"',"+qelement.name+")\n")
        for line in str(GMatch.script).split('\n'):
            scriptList.append('\t'+line+'\n')
        scriptList.append("def addToContext(k,v):\n")
        scriptList.append("\tdict[k]=v\n")
        scriptList.append("ret = script()")
        scriptFunction = string.join(scriptList,"")

        self.code = compile(str(scriptFunction), "GraphMatch script", "exec")


    def appliesTo(self, element):
        return self.types.exists(lambda t: element.oclIsKindOf(t))

    def search(self, targetElement):
        results = []
        
        if type(targetElement) == coral.kernel.ProjectPtr or type(targetElement) == coral.kernel.Project:
            targetElement = targetElement.contents()[0]

        if self.match.queryKind == 'csearch':
            for match in S.csearch(self.query, targetElement):
                try:
                    result = self.matchesElement(match)
                    if result:
                        results.append(result)
                    
                except Exception, exce:
                    print str(exce),self.elem

        else:
            for match in S.csearch_elem(self.query, targetElement):
                try:
                    result = self.matchesElement(match)
                    if result:
                        results.append(result)
                except Exception, exce:
                    print str(exce), self.elem
        return results


    def matchesElement(self, match):
        dict = {}
        locals = {"dict":dict, "match":match, "coral":coral, "ret":None}
        self.fillLocals(locals, match)
        ii = II(locals)

        ii.runcode(self.code)
        if ii.fail:
            raise Exception(string.join(ii.message))

        if locals["ret"]:

            return locals["dict"]
            
        return []
    
    def fillLocals(self, locals, match):
        for q_elem in self.query.queryelement:
            locals[str(q_elem.name)] = match[q_elem.abstractElement]


        
class ExpressionMatchEvaluator(MatchEvaluator):

    def __init__(self, EMatch):
        MatchEvaluator.__init__(self, EMatch)
        scriptList = []
        scriptList.append("import coral\ndef script():\n")
        scriptList.append("\taddToContext('self',self)\n")
        for line in str(EMatch.script).split('\n'):
            scriptList.append('\t'+line+'\n')
        scriptList.append("def addToContext(k,v):\n")
        scriptList.append("\tdict[k]=v\n")
        scriptList.append("ret = script()")
        scriptFunction = string.join(scriptList,"")
        self.code = compile(str(scriptFunction), "ExpressionMatch script", "exec")
        
        
    def search(self, target):
        results = []
        elem_kind = coral.AbstractElementDefinition_getAbstractElementDefinitionFromSMD(self.elem.kind)

        """ if target is a project """
        if type(target) == coral.kernel.ProjectPtr or type(target) == coral.kernel.Project:

            for match in target.contents():
                if match.oclIsKindOf(elem_kind):
                    result = self.matchesElement(match)
                    if result:
                        results.append(result)

        else:
            if self.elem.lookUp == 'allPartsAndMyself':
                for match in target.__getAllParts__():
                    if match.oclIsKindOf(elem_kind):
                        result = self.matchesElement(match)
                        if result:
                            results.append(result)

            elif self.elem.lookUp == 'allPartsNotMyself':
                space = target.__getAllParts__()
                space.drop(target)
                for match in space:
                    if match.oclIsKindOf(elem_kind):
                        result = self.matchesElement(match)
                        if result:
                            results.append(result)

            elif self.elem.lookUp == 'containedAndMyself':
                for match in target.__getAllParts__().select(lambda l:(l.__parent__() == target)or(l == target)):
                    if match.oclIsKindOf(elem_kind):
                        result = self.matchesElement(match)
                        if result:
                            results.append(result)

            elif self.elem.lookUp == 'containedNotMyself':
                for match in target.__getAllParts__().select(lambda l:l.__parent__() == target):
                    if match.oclIsKindOf(elem_kind):
                        result = self.matchesElement(match)
                        if result:
                            results.append(result)

            elif self.elem.lookUp == 'onlyMyself':
                if target.oclIsKindOf(elem_kind):
                    result = self.matchesElement(target)
                    if result:
                        results.append(result)
        return results


    def appliesTo(self, element):
        if not self.elem.kind:
            return False

        elem_kind = coral.AbstractElementDefinition_getAbstractElementDefinitionFromSMD(self.elem.kind)
        if element.oclIsKindOf(elem_kind):
            return True

        return False
        
    def matchesElement(self, match):
        
        dict = {}
        locals = {"dict":dict,"ret":None,"self":match}
        ii = II(locals)

        ii.runcode(self.code)
        if ii.fail:
            raise Exception(string.join(ii.message))

        if locals["ret"]:
            return locals["dict"]

        return []

    
