
# Interface functions: generate_code
#                      generate_code_elem
#                      save_to_disc

import coral
import os
from coral.addins.MATCH10.evaluate import *
from coral.modeler.extension import getFactory
from coral.util.airspeed.airspeed import Template
from coral.modeler.unicode import PyStringToUTF8
from coral.addins.TEXTAGEN10.proxy import Proxy

LM = coral.lang.MATCH10
LT = coral.lang.TEXTAGEN10



def generate_code(generatorFile, targetModelFile, outputDir="tmp"):
    """Generates text from a target model.
    Takes an XMI file containing a TEXTAGEN10::Package as first parameter.
    Takes an XMI file containing the target model as second parameter.
    Takes a string representing the output directory as third parameter (optional)."""

    global __engine__
    __engine__.result = {}

    if outputDir[-1] == os.sep:
        outputdir = str(outputDir)
    else:
        outputdir = str(outputDir + os.sep)
        
    loadGeneratorLanguageFromFile(generatorFile)
    loadTargetFromFile(targetModelFile)
    
    currentDirList = []
    currentDirList.append(outputdir)
    
    target = __engine__.getTargetProject()
    languagepackage = __engine__.getGeneratorLanguage()


    __engine__.generate(target, currentDirList, languagepackage.element)

    return __engine__.result




def generate_code_elem(textagenpackage, targetelement, outputDir="tmp"):
    """Generates text from a target model.
    Takes an element of type TEXTAGEN10::Package as first parameter.
    Takes an element of type CORE01::AnyElement as second parameter (target element).
    Takes a string representing the output directory as third parameter (optional)."""
    
    global __engine__
    assert textagenpackage.oclIsKindOf(LT.Package)
    assert targetelement.oclIsKindOf(coral.lang.CORE01.AnyElement)
    __engine__.result = {}
    
    if outputDir[-1] == os.sep:
        outputdir = str(outputDir)
    else:
        outputdir = str(outputDir + os.sep)

    currentDirList = []
    currentDirList.append(outputdir)

    __engine__.setGeneratorLanguage(textagenpackage)
    __engine__.setTargetModel(targetelement.__getRoot__())

    prj=coral.Project()
    targetelement.__setProject__(prj)
    
    __engine__.generate(prj, currentDirList, textagenpackage.element)

    return __engine__.result





def save_to_disc(result):
    """Writes the result of the generation to disc. Parameter: a
    result dictionary."""
    
    for k in result.keys():
        path = ""
        for folder in k.split(os.sep)[:-1]:
            path+=folder+os.sep
            __makeDir(path)
        dumpToFile(k, result[k])






class Engine:

    def __init__(self):
        self.targetProject = None
        self.targetModel = None
        self.generatorLanguage = None
        self.result = {} #File path and name as key, a list of file contents as value


    def setTargetProject(self, project):
        self.targetProject = project


    def getTargetProject(self):
        return self.targetProject


    def setTargetModel(self, tm):
        self.targetModel = tm


    def getTargetModel(self):
        return self.targetModel


    def setGeneratorLanguage(self, gl):
        assert gl.oclIsKindOf(LT.Package)
        self.generatorLanguage = gl


    def getGeneratorLanguage(self):
        return self.generatorLanguage


    def getMatchObject(self,match):
        assert match.oclIsKindOf(LM.Match)
        if match.oclIsKindOf(LM.ExpressionMatch):
            if not match.kind:
                raise Exception("ExpressionMatch without kind")
        c = getFactory(match.getImplementationPath()) 
        mObj = c(match)
        return mObj







    def generate(self, targetelement, curr, filesystemelements, firstrun=True):
        currentDirList = curr
        targetElement = targetelement
        for m in filesystemelements:
        
            if m.oclIsKindOf(LT.FolderGen):
                foldercontexts = checkSingleMatch(m.match, targetElement)
            
                for foco in foldercontexts:
                    foco["transform"] = Proxy()
                    # Setting foldername
                    if m.folderName:
                        template = Template(str(m.folderName))
                        currentFolder = str(template.merge(foco))
                        if not currentFolder[-1] == os.sep:
                            currentFolder += os.sep
                    else:
                        if m.match.oclIsKindOf(LM.GraphMatch):
                            currentFolder = foco[str(m.match.mainElement.name)].name + os.sep
                        else:
                            currentFolder = foco["self"].name + os.sep

                    currentDirList.append(currentFolder)

                    if currentDirList[-1] == currentDirList[-2] and not firstrun:
                        currentDirList.pop()
                        continue

                    pathAsString = ""
                    for e in currentDirList:
                        pathAsString += e

                    if m.subFolders: #generate sub-folders if current FolderGen element has any
                        if m.match.oclIsKindOf(LM.GraphMatch):
                            self.generate(foco[str(m.match.mainElement.name)], currentDirList, m.subFolders, False)
                        else:
                            self.generate(foco["self"], currentDirList, m.subFolders, False)
                    
                    if m.match.oclIsKindOf(LM.GraphMatch):
                        self.generate(foco[str(m.match.mainElement.name)], currentDirList, __engine__.getGeneratorLanguage().element, False)
                    else:
                        self.generate(foco["self"], currentDirList, __engine__.getGeneratorLanguage().element, False)
                    currentDirList.pop()

                    for f in m.subFiles:
                        if m.match.oclIsKindOf(LM.GraphMatch):
                            self.generatefiles(f, foco[str(m.match.mainElement.name)], pathAsString)
                        else:
                            self.generatefiles(f, foco["self"], pathAsString)
                        
            elif m.oclIsKindOf(LT.FileGen):
                # FileGen elements contained by a TEXTAGEN10.Package element must be matched as well, but only during the first iteration.
                if firstrun:
                    if not m.match.oclIsKindOf(LM.GraphMatch):
                        self.generatefiles(m, targetElement, currentDirList[0])
                    else:
                        self.generatefiles(m, __engine__.targetModel, currentDirList[0])




    def generatefiles(self, filegenelement, target, pathAsString):

        filecontexts = checkSingleMatch(filegenelement.match, target)
    
        for fico in filecontexts:
            fico["transform"] = Proxy()
            if not filegenelement.fileExtension:
                raise Exception("All FileGen element must have a fileExtension!")
            if filegenelement.fileExtension[0] == ".":
                extension = filegenelement.fileExtension
            else:
                extension = "."+filegenelement.fileExtension

            # Setting filename
            if filegenelement.fileName:
                template = Template(str(filegenelement.fileName))
                name = str(template.merge(fico))
                currentFilename = name+extension
            else:
                if filegenelement.match.oclIsKindOf(LM.GraphMatch):
                    currentFilename = fico[str(filegenelement.match.mainElement.name)].name+extension
                else:
                    currentFilename = fico["self"].name+extension
        
            fileContents=[]
        
            # Generate text for the file
            for t in filegenelement.text:

                # Use the mainElement as target if we used a graphmatch for filematching.
                if filegenelement.match.oclIsKindOf(LM.ExpressionMatch):
                    textcontexts = checkSingleMatch(t.match, fico["self"])
                else:
                    textcontexts = checkSingleMatch(t.match, fico[str(filegenelement.match.mainElement.name)])

                for teco in textcontexts:
                    teco["transform"] = Proxy()
                    contents = ""

                    if t.oclIsKindOf(LT.TemplateTextGen):
                        template = Template(str(t.template))
                        contents += str(template.merge(teco))

                    elif t.oclIsKindOf(LT.ExpressionTextGen):
                        contents += str(t.expression)

                    if t.autoIndent:
                        contents = contents.replace('\n','\n\t')
                        contents = '\t'+contents
                        nextindent = 2
                    else:
                        nextindent = 1
                    
                    fileContents.append(contents)

                    # Try to generate subtext
                    if t.match.oclIsKindOf(LM.ExpressionMatch):
                        fileContents = self.generatesubtext(teco["self"], t, fileContents, nextindent)
                    else:
                        fileContents = self.generatesubtext(teco[str(t.match.mainElement.name)], t, fileContents, nextindent)

            self.result[pathAsString+currentFilename] = fileContents

        

        


    def generatesubtext(self, targetelement, textgenelement, filecontents, indents=1):
        fc = filecontents

        if textgenelement.oclIsKindOf(LT.TemplateTextGen):
            expression = textgenelement.template
        else:
            expression = textgenelement.expression

        for sub_t in textgenelement.subText:

            if sub_t.autoIndent:
                nrOfIndents = indents
            else:
                nrOfIndents = indents-1
                
            if str(expression).find("["+sub_t.name+"]") != -1:
                textcontexts = checkSingleMatch(sub_t.match, targetelement)
            
                # If no matches are found for the subtext, remove '[Subtextname]'
                if not textcontexts: 
                    for j in range(len(fc)):
                        fc[j] = fc[j].replace('['+sub_t.name+']\n\t', "")
                        fc[j] = fc[j].replace('['+sub_t.name+']\n', "")
                        fc[j] = fc[j].replace('['+sub_t.name+']', "")

                iterations = 1
            
                for t_c in textcontexts:
                    t_c["transform"] = Proxy()
                    contents = ""
                    if sub_t.oclIsKindOf(LT.TemplateTextGen):
                        template = Template(str(sub_t.template))
                        contents += str(template.merge(t_c))

                    elif sub_t.oclIsKindOf(LT.ExpressionTextGen):
                        contents += str(sub_t.expression)

                                            
                    contents = contents.replace('\n','\n'+'\t'*nrOfIndents)
                    contents = '\t'*nrOfIndents+contents
                    
                    # If the last subtext-match is processed, replace '[Subtextname]' with the contents
                    if iterations == len(textcontexts): 
                        for i in range(len(fc)):
                            if not sub_t.autoIndent:
                                fc[i] = fc[i].replace('\t'*(nrOfIndents)+'['+sub_t.name+']', contents)
                            else:
                                fc[i] = fc[i].replace('\t'*(nrOfIndents-1)+'['+sub_t.name+']', contents)
                                
                    # If more subtext-matches still are to be processed, replace '[Subtextname]' with the contents, and add '[Subtextname]' on newline 
                    else:
                        
                        for i in range(len(fc)):
                            if not sub_t.autoIndent:
                                fc[i] = fc[i].replace('\t'*(nrOfIndents)+'['+sub_t.name+']', contents+'\n'+'\t'*(nrOfIndents)+'['+sub_t.name+']')
                            else:
                                fc[i] = fc[i].replace('\t'*(nrOfIndents-1)+'['+sub_t.name+']', contents+'\n'+'\t'*(nrOfIndents-1)+'['+sub_t.name+']')
                            
                    iterations += 1
                
                    # Generate subtext for the subtext
                    if sub_t.subText:
                        if sub_t.match.oclIsKindOf(LM.GraphMatch):
                            fc = self.generatesubtext(t_c[str(sub_t.match.mainElement.name)], sub_t, fc, nrOfIndents+1)
                        else:
                            fc = self.generatesubtext(t_c["self"], sub_t, fc, nrOfIndents+1)

        return fc


        
__engine__ = Engine()







def loadTargetFromFile(fileName, modelName=None):

    global __engine__
    prj = coral.loadDocument(fileName)
    __engine__.setTargetProject(prj)
    if modelName:
        for elem in prj.contents().select(lambda l:l.name == modelName):
            __engine__.setTargetModel(elem)
    else:
        for elem in prj.contents():
            if not elem.oclIsKindOf(coral.lang.XMIDI20.Diagram):
                __engine__.setTargetModel(elem)


def loadGeneratorLanguageFromFile(fileName):

    global __engine__
    prj = coral.loadDocument(fileName)
    for elem in prj.contents().select(lambda l:l.oclIsKindOf(LT.Package)):
        __engine__.setGeneratorLanguage(elem)



def checkSingleMatch(match, target):
    """Checks a single Match against a specific target"""
    global __engine__
    assert match.oclIsKindOf(LM.Match)

    try:
        matchObject = __engine__.getMatchObject(match)
        results = matchObject.search(target)
    except Exception, exce:
        print str(exce),"matchname:",match.name
        return []
    return results



def __makeDir(dir):
    """Makes a directory"""
    try:
        os.mkdir(str(dir))
    except OSError,e:
        if str(e).find("File exists:") != -1:
            pass
        else:
            raise


def dumpToFile(path_and_filename, filecontents):
    """Dumps a file and its contents to disc.
    Takes the path to the file as first parameter.
    Takes a list of filecontents as second parameter"""
    ff = open(path_and_filename,'w')
    try:
        for line in filecontents:
            line = line.replace("\\t","      ")
            line = line.replace("\\(","(")
            line = line.replace("\\)",")")
            for l in line.split("\\n"):
                ff.write(PyStringToUTF8(l)+"\n")
    finally:
        ff.close



    
