# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.

#
# UML14_to_SMD01 - In-memory UML 1.4 Model to SMD 0.1 Model Transformer
#
# SYNOPSIS
#
# UML14_to_SMD01 inputfile [ outputfile ]
#
# DESCRIPTION
#
# Simple script to convert a UML 1.4 Model to SMD 0.1 model.
# This is used because we don't have a SMD editor, while we do have a
# UML editor
#
#
# AUTHORS
#
# Tomas Lillqvist 2005
#
#
#

# This script converts UML14 into SMD01
# Model -> Language
# Class -> ElementDefinition
# Association -> Property
# Generalizations -> sub-, super-classes.
#
# UML Tagged Values with the name "documentaton" are converted into SMD documentation elements
#
# It works by right clicking on a UML14 Model and selecting convert in
# the UML2SMD contextmenu entry.  If there is an element in the model
# with the same name as an element in CORE01 it will use the CORE
# element in the language instead of creating a new ElementDefinition.
# Otherwise it wouldn't be possible to use the CORE types.

# The convert function takes as input a UML14.Model and returns a
# SMD01.Language. Very simple.

import coral
import sys

UML=coral.lang.UML14
SMD=coral.lang.SMD01

def pr_chain(e):
    l = coral.Element__getParentsAndMyself__(e)
    return "::".join( [ str(i.name) for i in l ] )

def __fix_uml2_info_in_datatype_names(model):
    """Fix attributes which point to several different String Datatypes."""
    all = model.__getAllParts__().select(lambda x: x.oclIsKindOf(UML.Attribute))

    the_string = UML.DataType(name = "CORE01::String", namespace = model)

    for a in all:
        if not a.type:
            continue
        n = str(a.type.name)
        if not n.startswith("String"):
            continue
        if "{ordered}" in n:
            a.ordering = "ordered"
        l = -1
        u = -1
        if "[0..*]" in n:
            l = 0
            u = -1
        elif "[1..*]" in n:
            l = 1
            u = -1
        elif "[*]" in n:
            l = 0
            u = -1
        if not a.multiplicity:
            a.multiplicity = UML.Multiplicity()
        if len(a.multiplicity.range) == 0:
            a.multiplicity.range.append(UML.MultiplicityRange())
        a.multiplicity.range[0].lower = l
        a.multiplicity.range[0].upper = u
        a.type.namespace = None
        a.type = the_string
        
def __fix_uml2_dt(e, overrides):
    if e.oclIsKindOf(UML.Namespace):
        for c in e.ownedElement:
            __fix_uml2_dt(c, overrides)

    if e.oclIsKindOf(UML.DataType):
        if e.name == "Boolean":
            e.name = "CORE01::Boolean"
        elif e.name == "String":
            e.name = "CORE01::String"
        elif e.name == "Integer":
            e.name = "CORE01::Integer"
        elif e.name == "UnlimitedNatural":
            e.name = "CORE01::Integer"

    if e.oclIsKindOf(UML.Class):
        is_enum = False
        if len(e.name) > 4 and str(e.name)[-4:] == "Kind":
            is_enum = True
        elif e.name == "InteractionOperator":
            is_enum = True
        elif e.name == "MessageSort":
            is_enum = True

        if is_enum:
            overrides[str(e.name)] = e
            # This is an enumeration
            s = UML.Stereotype(name = "enumeration")
            e.stereotype.add(s)

def __fix_uml2(model):
    """This is only useful for the UML2-as-UML14XMI11.xmi file."""
    if model.name != "Project":
        return

    __fix_uml2_info_in_datatype_names(model)

    overrides = {}
    __fix_uml2_dt(model, overrides)
    #
    # Some Attributes point to DataType elements, instead of
    # the correct Classes. Set their type correctly here,
    # from the information gathered in the previous step __fix_uml2_dt.
    #
    for e in model.__getAllParts__().select(lambda x: x.oclIsKindOf(UML.Attribute)):
        if e.type and overrides.has_key(str(e.type.name)):
            old = e.type
            e.type = overrides[str(e.type.name)]
            old.namespace = None
        if not e.type:
            p = e.__parent__()
            # Enumerations don't need a type...
            if p and p.oclIsKindOf(UML.Class) and p.stereotype.exists(lambda s: s.name == "enum" or s.name == "enumeration"):
                continue
            print "ERROR no type", pr_chain(e)
            continue
    #
    # Some multiplicity ranges are given as [ -1, -1 ].
    # Whiskey-Tango-Foxtrot?
    for e in model.__getAllParts__().select(lambda x: x.oclIsKindOf(UML.MultiplicityRange)):
        if e.lower < 0:
            e.lower = 0
        if e.upper < 1:
            e.upper = 1

def getTaggedValue(element, name,defaultValue=None):
    for tv in element.taggedValue:
        if str(tv.name)==name:
            return str(tv.dataValue)
    return defaultValue

def addDocumentation(smd_e,uml_e):
    """ UML Tagged Values with the name "documentaton" are converted into SMD documentation elements"""
    doc=getTaggedValue(uml_e,"documentation")
    if doc:
        d=SMD.Documentation()
        d.text.append(doc)
        smd_e.documentation.append(d)

    
def convert(model):
    element2ed={} # contains the mapping from UML to SMD
    lang=SMD.Language(name=model.name)
    addDocumentation(lang,model)
    lang.__setXMIid__(model.__XMIid__())
    lang.version=int(getTaggedValue(model,"version","1"))
    lang.revision=int(getTaggedValue(model,"revision","0"))
    element2ed[model]=lang
    lang.URInamespace=getTaggedValue(model, "URInamespace", "http://mde.abo.fi/"+str(model.name)+"/"+str(lang.version)+"."+str(lang.revision))

    __fix_uml2(model)

    transform_classes(model,lang, element2ed)
    transform_associations(model, lang, element2ed)
    transform_subsets(model, lang, element2ed)
    return lang

def fixProperty(P, p1, p2, element2ed):
    # Configures property P to reflect the attributes of association ends p1 and p2.
    # P gets:
    #  name of p2 or, if not set, p2.participant's lowercase name
    #  aggregation from p1
    #  ordered from p2
    #  multiplicity from p2

    P.name = p2.name
    P.__setXMIid__(p2.__XMIid__())
    P.element_kind = element2ed[p2.participant]

    if not element2ed.get(p1.participant,None):
        print p1
        print p1.participant

    element2ed[p1.participant].properties.append(P)

    if not p2.name:
        P.name = str(p2.participant.name).lower()

    # Sometimes p2.multiplicity is not there...
    if p2.multiplicity and p2.multiplicity.range:
        m = p2.multiplicity.range[0]
        P.multiplicity_upper = m.upper
        P.multiplicity_lower = m.lower

    if P.multiplicity_upper == 0:
        P.multiplicity_upper = 1

    if str(p2.ordering) == "ordered":
        P.isOrdered = True

    if str(p1.aggregation) == "composite":
        P.type = "composition"
        
        
def transform_classes(model, lang, element2ed):
    lang.isAbstract=model.isAbstract
    for elem in model.ownedElement:
        if elem.oclIsKindOf(UML.Class) or elem.oclIsKindOf(UML.DataType):
            pos=unicode(elem.name).find("::") # For example SMD01::, CORE01::
            if pos>0:
                if not element2ed.get(elem,None):
                    META=getattr(coral.meta, elem.name[:pos])
                    ed=META.elements[str(elem.name[pos+2:])]
                    element2ed[elem]=ed
                    continue
                    
            if elem.stereotype.exists(lambda s: s.name == "enum" or s.name == "enumeration"):
                ed=SMD.EnumerationDefinition(name=elem.name)
            else:
                ed=SMD.ElementDefinition(name=elem.name)
            addDocumentation(ed,elem)

            ed.__setXMIid__(elem.__XMIid__())
            lang.elements.append(ed)
            element2ed[elem]=ed
            ed.isAbstract=elem.isAbstract
                
            if elem.stereotype.exists(lambda s: str(s.name)=="root"):
                lang.root_ed=ed

            if not elem.generalization and ed.oclIsKindOf(SMD.ElementDefinition):
                #print "No superclass specified for ",elem
                #print "Setting CORE01::AnyElement as superclass"
                ed.superclasses.append(coral.meta.CORE01.AnyElement)
                
    for elem in model.ownedElement: # Recursively create contained Languages from Packages.
        if elem.oclIsKindOf(UML.Package):
            ed=SMD.Package(package=lang, name=elem.name)
            addDocumentation(ed,elem)
            ed.__setXMIid__(elem.__XMIid__())
            element2ed[elem]=ed
            r = ed.__getRoot__()
            ed.URInamespace=getTaggedValue(elem, "URInamespace", "http://mde.abo.fi/"+str(r.name)+"/"+str(r.version)+"."+str(r.revision) + "/" + ed.name)
        if elem.oclIsKindOf(UML.Namespace):
            transform_classes(elem, ed, element2ed)

def transform_associations(model, lang, element2ed):
    for elem in model.ownedElement: # After creaing all ElementDefinitions the properties are added.
        if elem.oclIsKindOf(UML.Generalization):
            if not elem.child or not elem.parent:
                s = pr_chain(elem)
                sys.stderr.write("WARNING! Bad Generalization " + s + "!\n")
                continue
            if not elem.child.oclIsKindOf(UML.Class) \
               or not elem.parent.oclIsKindOf(UML.Class):
                continue
            element2ed[elem.child].superclasses.append(element2ed[elem.parent])

        if elem.oclIsKindOf(UML.Association):
            end1 = elem.connection[0]
            end2 = elem.connection[1]

            # Create properties for the element definitions

            end1Prop=None
            end2Prop=None
            
            if end2.isNavigable == "true":
                # Properties can only be added to ElementDefinition, not to DatatypeDefinitions.
                if not element2ed[end1.participant].oclIsKindOf(SMD.DatatypeDefinition):
                    end1Prop = SMD.Property()
                    addDocumentation(end1Prop,end1)
                    fixProperty(end1Prop, end1, end2, element2ed)

            if end1.isNavigable == "true":
                # Properties can only be added to ElementDefinition, not to DatatypeDefinitions.
                if not element2ed[end2.participant].oclIsKindOf(SMD.DatatypeDefinition):
                    end2Prop = SMD.Property()
                    addDocumentation(end2Prop,end2)
                    fixProperty(end2Prop, end2, end1, element2ed)

            # Set opposite:Property
            if end1Prop and end2Prop:
                end1Prop.opposite = end2Prop
                end2Prop.opposite = end1Prop

            if end1Prop:
                element2ed[end1] = end1Prop
            if end2Prop:
                element2ed[end2] = end2Prop

            # Set type of property to composite if needed
            if end2Prop and str(end1.aggregation) == "composite":
                end2Prop.multiplicity_lower=0
                end2Prop.multiplicity_upper=1
                
            if end1Prop and str(end2.aggregation) == "composite":
                end1Prop.multiplicity_lower=0
                end1Prop.multiplicity_upper=1


            #Set multiplicity for EnumerationDefinition properties to 1..1
            if end1Prop and element2ed[end2.participant].oclIsKindOf(SMD.EnumerationDefinition):
                end1Prop.multiplicity_lower=1
                end1Prop.multiplicity_upper=1

            if end2Prop and element2ed[end1.participant].oclIsKindOf(SMD.EnumerationDefinition):
                end2Prop.multiplicity_lower=1
                end2Prop.multiplicity_upper=1
                

        if elem.oclIsKindOf(UML.Class):
            # Attributes become composite properties with upper multiplicity 1
            for attr in elem.feature.select(lambda f: f.oclIsKindOf(UML.Attribute)):
                if elem.stereotype.exists(lambda s: s.name == "enum" or s.name == "enumeration"):
                    element2ed[elem].values.append(unicode(attr.name))
                    continue

                attr_ed=element2ed.get(attr.type, None)
                if not attr_ed:
                    pos=unicode(elem.name).find("::") # For example SMD01::, CORE01::
                    if pos>0:
                        META=getattr(coral.meta, elem.name[:pos])
                        ed=META.elements[str(elem.name[pos+2:])]
                        element2ed[elem]=ed
                    
                    else:
                        if not attr.type:
                            print "ERROR: No type definition for", pr_chain(attr), ", skipping!"
                            
                            continue
                        attr_ed=SMD.ElementDefinition(name=attr.type.name)
                        addDocumentation(attr_ed,attr)
                        element2ed[attr.type]=attr_ed

                newprop=SMD.Property(name=attr.name)
                addDocumentation(newprop,attr)
                newprop.__setXMIid__(attr.__XMIid__())
                newprop.element_kind=attr_ed

                if attr.ordering=="ordered":
                    newprop.isOrdered = True
                    
                m = attr.multiplicity.range
                if m:
                    newprop.multiplicity_lower = m[0].lower
                    newprop.multiplicity_upper = m[0].upper
                else:
                    newprop.multiplicity_lower = 1
                    newprop.multiplicity_upper = 1

                newprop.type = "composition"
                element2ed[elem].properties.append(newprop)

                if attr.initialValue and attr.initialValue.body:
                    newprop.defaultValue = attr.initialValue.body
                    
            # Operations have to be written in Python
            # All code goes into the specification slot of UML14 Operation.
            # UML14 Parameters are not used.
            for oper in elem.feature.select(lambda f: f.oclIsKindOf(UML.Operation)):
                newOper=SMD.Operation(name=oper.name)
                addDocumentation(newOper,oper)
                for line in unicode(oper.specification).split('\n'):
                    newOper.code.append(line)

                newOper.language="python"
                newOper.__setXMIid__(oper.__XMIid__())
                element2ed[elem].operations.append(newOper)

    for elem in model.ownedElement: # Recursively create contained associations
        if elem.oclIsKindOf(UML.Namespace):
            transform_associations(elem, element2ed[elem], element2ed)

def __uml14_is_subclassof(child, parent):
    """Returns true if child is a transitive subclass
    of parent in a UML 1.4 Model"""
    for i in child.generalization.parent:
        if i == parent:
            return True
        if __uml14_is_subclassof(i, parent):
            return True
    return False

def transform_subsets(model, lang, element2ed):
    for elem in model.ownedElement: # After creaing all ElementDefinitions the properties are added.
        if elem.oclIsKindOf(UML.Generalization):
            if not elem.child or not elem.parent:
                s = pr_chain(elem)
                sys.stderr.write("WARNING! Bad Generalization " + s + "!\n")
                continue
            if not elem.child.oclIsKindOf(UML.Association) \
               or not elem.parent.oclIsKindOf(UML.Association):
                continue
            p_ass = elem.parent
            c_ass = elem.child
            c0 = c_ass.connection[0]
            c1 = c_ass.connection[1]
            p0 = p_ass.connection[0]
            p1 = p_ass.connection[1]
            # Check if we need to swap p0,p1
            # This heuristic works almost always
            if not __uml14_is_subclassof(c0.participant, p0.participant):
                p0, p1 = p1, p0
            if not __uml14_is_subclassof(c0.participant, p0.participant):
                s = pr_chain(elem)
                sys.stderr.write("WARNING! Bad Generalization " + s + "!\n")
                continue
            element2ed[c0].supersets.append(element2ed[p0])
            element2ed[c1].supersets.append(element2ed[p1])
    for elem in model.ownedElement: # Recurse
        if elem.oclIsKindOf(UML.Namespace):
            transform_subsets(elem, element2ed[elem], element2ed)

def main(arguments):
    f = arguments[0]

    L = coral.lang.UML14
    p = coral.loadDocument(f)
    m_uml = None
    for i in p.contents():
        if i.oclIsKindOf(L.Model):
            m_uml = i
            break
    assert m_uml

    m_smd = convert(m_uml)
    if len(arguments) > 1:
        outfile = arguments[1]
    else:
        outfile = "/dev/stdout"

    prj = coral.Project()
    prj.setFilename(outfile)
    m_smd.__setProject__(prj)
    prj.setIOFormat(p.getIOFormat())
    coral.saveDocument(prj)

if __name__=='__main__':
    #
    # Set the user's locale here. Since Coral can be used as a library,
    # we do not tamper with the locale at all except at the
    # __name__ == "__main__" situations
    # The filesystem encoding is set automatically by Python
    # (and it cannot be changed after startup)
    #
    import locale
    locale.setlocale(locale.LC_ALL, '')
    import sys
    main(sys.argv[1:])
