#!/usr/bin/env python
# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.
#
# -*- coding: UTF-8 -*-
#
# smd2ecore - script to convert SMD metamodels to ECORE metamodels
#
# SYNOPSIS
#
# Explained in usage()
#
# DESCRIPTION
#
# smd2ecore is a script for converting Coral's SMD metamodels
# to ECORE metamodels (i.e. Eclipse).
#
# BUGS
#
# Probably, the ecore metamodel seems to be changing
#
# AUTHORS
#
# Marcus Alanen <maalanen@abo.fi> 2006
#

coral = None
core = None
ecore = None
ecore_model = None
smd = None
special_smd2ecore = None

def create(s):
    global special_smd2ecore
    if not special_smd2ecore:
        # BUG add more stuff here when you get unknown.xml#// errors!
        special_smd2ecore = { core.coralPyObj: None,
                              core.String: ecore_model.EString,
                              core.Name: ecore_model.EString,
                              core.Float: ecore_model.EFloat,
                              core.Boolean: ecore_model.EBoolean,
                              core.Integer: ecore_model.EInt,
                              core.UnlimitedInteger: ecore_model.EInt,
                              core.UnlimitedReal: ecore_model.EDouble,
                              core.Double: ecore_model.EDouble,
                              core.AnyElement: ecore_model.EObject,
                              }
    if special_smd2ecore.has_key(s):
        return special_smd2ecore[s]


    e = None
    if s.oclIsKindOf(smd.Package):
        e = ecore.EPackage()
        e.name = s.name
        e.nsURI = s.URInamespace
    elif s.oclIsKindOf(smd.EnumerationDefinition):
        e = ecore.EEnum()
        e.name = s.name
        for sv in s.values:
            el = ecore.EEnumLiteral()
            el.name = unicode(str)
            e.eLiterals.append(el)
    elif s.oclIsKindOf(smd.DatatypeDefinition):
        print s
        assert 0
    elif s.oclIsKindOf(smd.ElementDefinition):
        e = ecore.EClass()
        e.interface = s.isAbstract
        e.name = s.name
    elif s.oclIsKindOf(smd.Property):
        e = ecore.EReference()
        e.name = s.name
        if s.type == "composition":
            e.containment = "true"
        e.derived = s.isDerived
        e.ordered = s.isOrdered
        e.transient = s.isUnserializable
        e.upperBound = s.multiplicity_upper
        e.lowerBound = s.multiplicity_lower
        e.unique = s.isBag
        e.defaultValueLiteral = s.defaultValue
        for ss in s.supersets:
            a = ecore.EAnnotation()
            a.source = "subsets"
            # BUG fix a.references
            e.eAnnotations.add(a)
            
    return e

def connect(s2e):
    for s, e in s2e.iteritems():
        if e.__getRoot__() == ecore_model:
            continue
        if s.__getRoot__() == core:
            continue
        if s.oclIsKindOf(smd.Package):
            for child in s.elements:
                if s2e.has_key(child):
                    e.eClassifiers.add(s2e[child])
        if s.oclIsKindOf(smd.ElementDefinition):
            # Properties
            for child in s.properties:
                e.eStructuralFeatures.add(s2e[child])
            # BUG supersets
            # Superclasses
            for superclass in s.superclasses:
                e.eSuperTypes.add(s2e[superclass])
        if s.oclIsKindOf(smd.Property):
            if s.opposite and not s.opposite.isAnonymous:
                e.eOpposite = s2e[s.opposite]
            e.eType = s2e[s.element_kind]
            
def main(arguments):
    global ecore
    global ecore_model
    global smd
    global core
    global coral

    def usage():
        print "Usage: smd2ecore [ --help | --version | [ -o ecorefile ] smdfile"
        print "Options:"
        print "  --help         Display this information"
        print "  --version      Display version information and exit"
        print "  -o             Output file name for the ECORE model"
        print "If no output file is given, prints XMI file to stdout."
        print ""

    def version():
        import coral
        print "This is Coral version", coral.version, " (smd2ecore script)"
        
    import sys
    import getopt

    try:
        opts, args = getopt.getopt(arguments, "ho:", ["help", "version"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    opt_o = ""
    
    for o,a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif o in ("--version"):
            version()
            sys.exit(0)
        elif o in ("-o"):
            opt_o = a
        else:
            usage()
            sys.exit(2)

    if not args:
        usage()
        sys.exit(2)

    import coral
    import sys
    from coral.modeler.diagram.di_support import getAllLinkedProjects

    ecore = coral.lang.ecore10

    ecore_model = coral.loadDocument("coral/lang/ecore10/def/ecore_model.xmi").contents()[0]

    smd = coral.lang.SMD01

    pcore = coral.loadDocument("coral/core/CORE01.xmi")
    core = pcore.contents()[0]

    s2e = {}
    project = coral.loadDocument(args[0])
    allprojects = getAllLinkedProjects([project])
    
    root_class_in_smd = None
    for prj in allprojects:
        for root in prj.contents():
            for s in root.__getAllParts__():
                e = create(s)
                if e:
                    s2e[s] = e

    connect(s2e)

    p = coral.Project()
    for s, e in s2e.iteritems():
        if s.oclIsKindOf(smd.Language) and s.__getRootsProject__() == project:
            e.__setProject__(p)

    if opt_o:
        p.setFilename(opt_o)
        coral.saveDocument(p)
    else:
        s = coral.saveDocument_str(p.contents(), "XMI 2.0")
        print s

if __name__ == "__main__":
    #
    # Set the user's locale here. Since Coral can be used as a library,
    # we do not tamper with the locale at all except at the
    # __name__ == "__main__" situations
    # The filesystem encoding is set automatically by Python
    # (and it cannot be changed after startup)
    #
    import locale
    locale.setlocale(locale.LC_ALL, '')

    import sys
    main(sys.argv[1:])
