# This file is part of the Coral project.
# Please read the copyright notice in the LICENSE file.

import traceback
from coral.modeler.editor import Editor,EditorSet
from coral.modeler.window import CoralMDIWindow
from coral import *
from qt import *
from coral.modeler.getFQName import *

from coral import modeler
from coral.modeler import iconcache
from coral import log
from coral import configuration
from coral.modeler import extension
from coral.modeler.modeler import CoralModeler

import types

logger=log.getLogger("ListEditor")

def metamodelName(element):
    l=element.__type__().__language__()
    return l.getName()+unicode(l.getVersion())+unicode(l.getRevision())

#
# Amount of list items created. Used in unittests.
#
DEBUG_N_LIST_ITEMS = 0

class ListItem(QListViewItem):
    def __init__(self, parent, subject):
        global DEBUG_N_LIST_ITEMS
        DEBUG_N_LIST_ITEMS += 1
        QListViewItem.__init__(self, parent)
        self.subject = subject
        self.onElementChanged()
        self.setOpen(0)
        
    def setOpen(self, b):
        """This is called when the node is first expanded.
        We use it to populate ourselves."""
        QListViewItem.setOpen(self, b)
        if b:
            self.updateChildren()

    def updateChildren(self):
        pass

class ElementListItem(ListItem):
    def __init__(self, parent, subject):
        ListItem.__init__(self, parent, subject)

        typename = self.subject.__type__().getName().lower()
        icon=iconcache.getIcon(typename + "lv", metamodelName(self.subject) )
        if icon:
            self.setPixmap(0,icon)

    def updateChildren(self):
        editor = self.listView()

        children={}
        
        for p in self.subject.__type__().__properties__():
            if not p.isComposition():
                continue
            s = self.subject.__getSlotRO__(p.getName())
            if not s:
                continue
            for i in s:
                if editor.qlvItems.has_key(i):
                    item = editor.qlvItems[i]
                    self.insertItem(item)
                    children[item] = 1
                else:
                    if isinstance(i,Element):
                        an_item = editor.addElementItem(i,self)
                        if not an_item:
                            continue
                        children[an_item] = 1

        # remove the ListItem children that should not be there
        i = self.firstChild()
        while i:
            next = i.nextSibling()
            if not children.has_key(i):
                i.__kill__()
            i = next

    def onElementChanged(self):
        if not self.subject:
            return

        if not self.subject.__getRootsProject__():
            self.__kill__()
            return

        editor = self.listView()
        if not editor:
            return

        # Explicitly call it so that our view of our children is updated
        if self.isOpen():
            self.updateChildren()

        name = getFQName(self.subject, 1)
        if not name:
            name="(noname)"
        else:
            # Must flatten name because QT can't show multi-line strings
            # in ListViews
            name = re.sub("\n", " ", name)

        # Prepend the name of the property owning the element, if exists
        if self.subject.__parentSlot__() and self.subject.__parentSlot__():
            name=unicode(self.subject.__parentSlot__().__type__().getOppositeProperty().getName())+" "+name
            
        self.setText(0,QString(name+":"+self.subject.__type__().getName()))

        if editor.pickMode and editor.includeCriteria:
            selectable=0
            for c in editor.includeCriteria:
                if c(self.subject):
                    selectable=1
                    break
            self.setSelectable(selectable)
                
    def checkExpandMark(self, outliner):
        """ Return True if we have even one child which we can
        show as a child"""
        if not self.subject:
            return False
        for s in self.subject.__all_slots__():
            p = s.getProperty()
            if not p.isComposition():
                continue
            # primitives not shown
            if p.getTarget().isDatatypeDefinition():
                continue
            if not len(s):
                continue
            for element in s:
                b = outliner.checkType(element)
                if b:
                    return True
        return False
    
            
    def __kill__(self):
        logger.info("__kill__"+unicode(self)+" "+unicode(self.parent()))
        p = self.parent()
        if not p:
            p = self.listView()
        if p:
            p.takeItem(self)

class ProjectListItem(ListItem):
    def updateChildren(self):
        children={}
        for i in self.subject.contents():
            editor = self.listView()
            if editor.qlvItems.has_key(i):
                item = editor.qlvItems[i]
                self.insertItem(item)
                children[item] = 1
            else:
                if isinstance(i,Element):
                    an_item = editor.addElementItem(i,self)
                    if not an_item:
                        continue 
                    children[an_item] = 1

        # remove the ListItem children that should not be there
        i = self.firstChild()
        while i:
            next = i.nextSibling()
            if not children.has_key(i):
                i.__kill__()
            i = next

    def onElementChanged(self):
        if not self.subject:
            return


        if self.subject.isInvalid():
            self.__kill__()
            return
        
        if self.isOpen():
            self.updateChildren()
       
        icon=iconcache.getIcon("storagelv" )
        if icon:
            self.setPixmap(0,icon)

        name = self.subject.getHrefFilename()
            
        if not name:
            name="(noname)"
        if self.subject.isTransientProject():
            name += " <<transient>>"
        self.setText(0,QString(name+":Project"))

        editor = self.listView()
        self.setSelectable(0)
                                    
    def checkExpandMark(self, outliner):
        return True
    
    def __kill__(self):
        self.listView().removeItem(self)
        
class ElementOutlinePick(QDialog):
    # This used to be the ElementPick
    def __init__(self,parent=None):
        QDialog.__init__(self,parent)      
        self.layout = QGridLayout(self,1,1,0,6,"Layout")
        
        
    def pick(self, element, theType):
        list=OutlineEditor(self,includeCriteria=[lambda e:e.oclIsKindOf(theType)])
        self.layout.addWidget(list,0,0)
        list.pickMode=1
        if element:
            list.focusOn(element)

        self.show()
        self.exec_loop()
        r=list.getSelectedElement()
        self.removeChild(list)
        list.destroy()
        list=None
        return r

class ElementPick(QDialog):
    # This is the new version of ElementPick
    # The old version is renamed to ElementOutlinePick
    def __init__(self, editorSet=None, parent=None):
        QDialog.__init__(self,parent,"Element Pick",True)
        if editorSet is None:
            editorset = EditorSet()
        self.editorSet = editorSet
        self.setCaption("Select an Element")

        self.tabWidget = QTabWidget(self,"TabWidget")

        self.bFrame = QFrame(self,'bFrame')
        self.bFrame.setMinimumSize(QSize(30,30))
        self.bFrame.setFrameShape(QFrame.StyledPanel)
        self.bFrame.setFrameShadow(QFrame.Raised)
        self.bFrame.setLineWidth(0)
        
        self.openButton = QPushButton(self.bFrame,"openButton")
        self.openButton.setIconSet(QIconSet(iconcache.getIcon("fileopen")))
        self.openButton.setGeometry(QRect(0,0,30,30))
        self.openButton.show()
        
        self.openButton.connect(self.openButton,SIGNAL("clicked()"),self.openProject)
         
        layout = QGridLayout(self,1,1,0,6,"Layout")
        layout.addWidget(self.bFrame,0,0)  
        layout.addWidget(self.tabWidget,1,0)

        self.itemToElement={}
        self.selected = None
        
    def setupOutline(self, element, theType):
        self.outline=OutlineEditor(self.editorSet,self,includeCriteria=[lambda e:e.oclIsKindOf(theType)])
        self.tabWidget.insertTab(self.outline,QString.fromLatin1("outline"))

        self.outline.pickMode=1
        if element:
            self.outline.focusOn(element)

        return self.outline

    def setupElementListPick(self, element):
        self.elementListPick = QListView(self)
        self.elementListPick.addColumn("Element")
        self.elementListPick.addColumn("Location")
        self.tabWidget.insertTab(self.elementListPick,QString.fromLatin1("list"))
        self.connect(self.elementListPick, SIGNAL("clicked(QListViewItem *)"),
                     self.clicked)
        self.populatePickList()
        return self.elementListPick
    
    def populatePickList(self):
        self.elementListPick.clear()
        for project in coral.getAllProjects():
            projectName=project.getFilename().split("/")[-1]
            for root in project.contents():
                for part in root.__getAllParts__():
                    if part.oclIsKindOf(self.theType):
                        item = QListViewItem(self.elementListPick)

                        name = getFQName(part, 1)
                        if not name:
                            name="(noname)"
                        item.setText(0, name)

                        location=""
                        e=part
                        while e.__parent__():
                            e=e.__parent__()
                            location="::"+getFQName(e,1)+location
                        location = "("+projectName+")"+location
                        item.setText(1, location)
                        self.itemToElement[item]=part

        return self.elementListPick

    def openProject(self):
        import coral.modeler.project
        coral.modeler.project.openProject()
        self.populatePickList()
        
    def pick(self, element, theType):
        self.theType=theType
        self.outline = self.setupOutline(element, theType)
        self.elementListPick = self.setupElementListPick(element)
        
        self.show()
        self.exec_loop()

        r = self.selected
        if not r:
            r=self.outline.getSelectedElement()

        self.removeChild(self.outline)
        self.removeChild(self.elementListPick)

        self.outline.close(True)
        self.outline=None

        self.elementListPick.destroy()
        self.elementListPick=None

        return r

    def clicked(self, item):
        # This gets called if an item in the elementListPick is clicked, but not when items in the outline are selected
        # Outline handles itself, 
        if not item:
            return
        self.selected = self.itemToElement[item]
        self.done(1)


class OutlineEditor(Editor,CoralMDIWindow,QListView):

    def __init__(self,
                 editorSet,
                 parent,
                 excludeCriteria=None,includeCriteria=None):
        Editor.__init__(self, editorSet)
        CoralMDIWindow.__init__(self, parent)
        QListView.__init__(self, parent)

        self.setCaption("Model Outline")
        self.pickMode=0
        self.showInternalProjects=False
        
        if includeCriteria:
            self.includeCriteria=includeCriteria
        else:
            self.includeCriteria=[]
           
        if excludeCriteria:
            self.excludeCriteria=excludeCriteria
        else:
            self.excludeCriteria=[]
        
        self.setRootIsDecorated(1)
        self.setTreeStepSize(12)
        self.addColumn("Element")
        #self.addColumn("Type")

        self.qlvItems={}
        self.connect(self,SIGNAL("clicked(QListViewItem *)"),
                     self.clicked)
     
  
        self.connect(self,SIGNAL("rightButtonPressed(QListViewItem *,const QPoint &, int)"),
                     self.rightClicked)

        self.result=None

        self.contextMenuExtensions=[]


        for dtype,entry in CoralModeler.extensionRegistry.getRegisteredExtensionsByType(type="ContextMenuAction"):
            self.contextMenuExtensions.append(entry(self.editorSet))

        getTransactionManager().addObserver(self.onModelChanged)
        
        for s in coral.getAllProjects():
            self.addProject(s)

    def startDrag(self):
        
        item = self.currentItem()
        if item:
            modelitem = item.subject
            if modelitem and not isinstance(modelitem, coral.Project):
                if modelitem.oclIsKindOf(AnyElement):
                    dobj = coral.lang.CoralEx10.InternalQtDragAndDrop.DragAndDrop()
                    dobj.linkedContents = [modelitem]
                    o = coral.Set()
                    o.add(dobj)
                    xmi_str = coral.saveDocument_str(o, "XMI 2.0")
                    d =  QTextDrag(QString(xmi_str), self);
                    logger.debug("\nDrag & Drop started. Output:")
                    d.dragCopy()

    def contentsMouseMoveEvent(self,e):
        if e.state() & Qt.LeftButton or e.state() & Qt.RightButton :
            ## Initiate drag procedure
            self.startDrag()


    def setIncludeCriteria( self, includeCriteria=None ):
        if includeCriteria:
            self.includeCriteria=includeCriteria
        else:
            self.includeCriteria=[]
           

    def setExcludeCriteria( self, excludeCriteria=None ):
        if excludeCriteria:
            self.excludeCriteria=excludeCriteria
        else:
            self.excludeCriteria=[]

    def close(self, alsoDelete):
        getTransactionManager().removeObserver(self.onModelChanged)
        self.editorSet.unlink(self)
        return QListView.close(self, alsoDelete)

    def onModelChanged(self,x):
        logger.debug("Begin OutlineEditor.onModelChanged")
        updated={}
        tm=getTransactionManager()
        tr=tm.getCurrentTransaction()

        # ElementItems added this round.
        # Used so we don't add a whole chain of new items,
        # only the first one
        recently_added = {}

        for c in tr:
            if isinstance(c,ElementChangeSlotCommand):
                element=c.getSlot().getElement()

                if element in self.qlvItems:
                    if element not in updated:
                        item = self.qlvItems[element]
                        item.onElementChanged()
                        self.__setExpandMark(item)
                        updated[element]=1
                         
                else:
                    # Only the first new child is added
                    # This is ok.
                    parent = element.__parent__()
                    if parent in self.qlvItems and self.qlvItems[parent].isOpen():
                        if parent in recently_added:
                            continue
                        recently_added[parent] = True
                        if not parent.__getRootsProject__():
                            continue
                        self.addElementItem(element, self.qlvItems[parent])
                        updated[element] = 1
            elif isinstance(c, ProjectNameChangeCommand):
                project=c.getProject()
                item = self.qlvItems.get(project, None)
                if item:
                    item.onElementChanged()
                updated[project]=1
            elif isinstance(c, coral.NewProjectCommand):
                element = c.getNewProject()
                self.addProject(element)
            elif isinstance(c, coral.ProjectChangeCommand):
                element = c.getAtom()
                oldproject = c.getOldProject()
                # HACK! We don't have a ProjectInvalitedCommand
                if oldproject and oldproject.isInvalid():
                    item = self.qlvItems.get(oldproject,None)
                    if item:
                        self.removeItem(item)
                if element:
                    if oldproject: 
                        item = self.qlvItems.get(oldproject,None)
                        if item:
                            item.onElementChanged()
                    updated[oldproject] = 1
                    the_project = element.__getProject__()
                    if the_project:
                        item = self.qlvItems.get(the_project,None)
                        if item:
                            item.onElementChanged()
                    updated[the_project] = 1

        logger.debug("End OutlineEditor.onModelChanged")
                    
    def focusOn(self, element):
        if not self.checkType(element):
            return None
        logger.debug("OutlineEditor.focusOn" + unicode(self) + " " + unicode(element))
        self.focus=element

        self.clearSelection()

        if not self.qlvItems.has_key(element):
            # This handles the case where the user activates an element
            # whose parents are not in the outline editor.
            previous_parent = self
            for p in coral.Element__getParentsAndMyself__(element):
                par = p.__parent__()
                if not par:
                    storage=p.__getProject__()
                    if not storage:
                        continue # ??
                    previous_parent = self.addProject(storage,force=1)            
                else:
                    if self.checkType(par):
                        if not self.qlvItems.has_key(par):
                            previous_parent=self.addElement(par)
                        else:
                            previous_parent=self.qlvItems[par]
                if not self.qlvItems.has_key(p):
                    self.addElementItem(p, previous_parent)


        p = self.qlvItems[element]

        self.setSelected(p,1)

        self.ensureItemVisible(p)

        
    def addElement(self,element):
        return self.addElementItem(element,self)

    def addProject(self,storage,force=0):
        if storage.isInvalid():
            return
        if self.qlvItems.has_key(storage):
            return self.qlvItems[storage]

        if storage.isUserProject() or self.showInternalProjects or force:
            item = ProjectListItem(self, storage)
            self.__setExpandMark(item)
            self.qlvItems[storage] = item
            assert self.qlvItems.has_key(storage)
            return item
        return None
    
    def checkType(self,element):
        """ returns true if the type element is compatible with this view
        If the type of element is in excludeTypes it will not be shown.       
        If includeTypes is defined, then an element will be show if its type is in includeTypes"""

        if isinstance(element,coral.Project):
            return 1
        
        if not self.includeCriteria and not self.excludeCriteria:
            return 1
            
        for c in self.excludeCriteria:
            if c(element):
                return 0
       
        if not self.includeCriteria or self.pickMode:
            return 1
        else:        
            for c in self.includeCriteria:
                if c(element):
                    return 1
                #if element.oclIsKindOf(t):
                #    return 1                
            return 0
        
   
    
    def __setExpandMark(self, item):
        b = item.checkExpandMark(self)
        if b or ( not b and item.isExpandable() ):
            item.setExpandable(b)
            
    def addElementItem(self,element,parent):
        """Add a new ListItem for an element to parent, if it passes
        the matching criteria."""
        if not self.checkType(element):
            return None
        if self.qlvItems.has_key(element):
            return self.qlvItems[element]
        item = ElementListItem(parent, element)
        self.__setExpandMark(item)
        self.qlvItems[element] = item
        return item

    def __removeItem(self, item):
        i = item.firstChild()
        while i:
            next = i.nextSibling()
            self.__removeItem(i)
            i = next
        del self.qlvItems[item.subject]
        # Qt oddity, can't call parent() on top-most elements
        p = item.parent()
        if not p:
            p = item.listView()
        p.takeItem(item)
        
    def removeItem(self, item):
        """Removes an item from a ListView. The given parameter
        cannot be used after this. Especially note how this means
        you have to save item.nextSibling() before calling this method."""
        # Hm, could we use a QListViewItemIterator?
        self.__removeItem(item)
        
    def rightClicked(self, item, pos, c):
        # Need to be able to show/hide even when in pickmode!
        ###if self.pickMode:
        ###    return
        
        contextMenu=QPopupMenu(self,"cm")

        if item:
            for o in self.contextMenuExtensions:
                o.createMenuItem(contextMenu,[item.subject],[])
        if not item or isinstance(item.subject, coral.Project):
            contextMenu.insertSeparator()
            if self.showInternalProjects:
                verb="Hide"
            else:
                verb="Show"
                
            contextMenu.insertItem(verb+" System Projects",self.showHideInternalProjects)
        contextMenu.popup(QPoint(pos.x()+5,pos.y()+10),0)

    def showHideInternalProjects(self,a):
        self.showInternalProjects=not self.showInternalProjects
        self.clear()
        self.qlvItems={}
        for s in coral.getAllProjects():
            self.addProject(s)

    def clicked(self,item):
        if not item:
            return

        if item.isOpen():
            item.updateChildren()
        if self.pickMode and item.isSelectable():
            self.done(1)
        else:
            if not isinstance(item.subject,Project):
                self.notifyFocusOn(item.subject)
        
    def keyPressEvent(self,e):
        QListView.keyPressEvent(self,e)
        logger.debug(unicode(e.key()))
        if e.key()==Qt.Key_Return:
            item=self.selectedItem()
            if item and item.isSelectable():
                if self.pickMode:
                    self.done(1)
        
##     def contentsMouseDoubleClickEvent(self,e):
##         pass
    
##     def contentsMouseMoveEvent(self,e):
##         pass

    def apply(self):
        if self.firstChild():
            self.firstChild().updateChildren()

    def done(self,x):
        if self.selectedItem():
            self.result=self.selectedItem().subject
        else:
            self.result=None

        # Incase the outline is contained in a QTabWidget the parent
        # cannot provide the done method

        p=self.parent()
        while not hasattr(p,"done"):
            p=p.parent()
            
        return p.done(x)
        
    def getSelectedElement(self):
        return self.result

    
    
if __name__=='__main__':
    #
    # Set the user's locale here. Since Coral can be used as a library,
    # we do not tamper with the locale at all except at the
    # __name__ == "__main__" situations
    # The filesystem encoding is set automatically by Python
    # (and it cannot be changed after startup)
    #
    import locale
    locale.setlocale(locale.LC_ALL, '')

    a=QApplication([])
    model=loadModel("coral/modeler/outline/unittests/model.xmi")
    r=ElementPick(EditorSet()).pick(model, coral.lang.UML14.Class)
    logger.debug("Selected "+unicode(r)+" "+unicode(r.name))
