import coral
import copy
import time

import util as U
from planning import SearchPlan

class MatchData:

    """class MatchData(query_set, **options):
    The options are:
    \tallow_permutations = Boolean (default False)
    \tcmp_elem_fun = Pointer to function used to compare element types
    \t\t(default coral.util.search.util.isKindOf)
    \tstate_T = State class which implements visitor (QueryState by default)
    
    query_set should be of type coral.Set and it must contain every element
    \tfor which a matching is wanted."""

    def __init__(self, query_set, **options):
        assert isinstance(query_set, coral.Set)
        self.elem_cmp_fun={}
        self.disjoint_elements=coral.Set()
        self.active_elements=coral.Set()
        self.negative_elements=coral.Set()
        
        self.query_set=query_set

        # Count elements in query_set
        self.query_size = len(query_set)

        # Configurable things
        self.allow_permutations = False
        self.cmp_elem_fun       = U.isKindOf # isSameType
        self.state_T            = QueryState # implements the visitor

        for k,v in options.items():
            if hasattr(self, k):
                assert type(getattr(self,k)) == type(v)
                setattr(self, k, v)
            else:
                assert 0, "no such option"

        self.elem_property_lists=self.__make_property_lists()

    def __make_property_lists(self):
        elem_property_lists={}
        for elem in self.query_set:

            Elem_and_ElemPrim_lists=[{},{}]
            elem_property_lists[elem]=Elem_and_ElemPrim_lists

            for slot in elem.__all_slots__():
                prop = slot.getProperty()
                
                if prop.getName() == "xmi_extension":
                    continue

                if self.__isPrimitive(prop):
                    Elem_and_ElemPrim_lists[1][prop]=len(slot)
                else:
                    if len(slot) == 0:
                        continue

                    Elem_and_ElemPrim_lists[0][prop]=len(filter(lambda x: x in self.query_set,slot))

        for elem in self.negative_elements:

            Elem_and_ElemPrim_lists=[{},{}]
            elem_property_lists[elem]=Elem_and_ElemPrim_lists

            for slot in elem.__all_slots__():
                prop = slot.getProperty()
                
                if prop.getName() == "xmi_extension":
                    continue

                if self.__isPrimitive(prop):
                    Elem_and_ElemPrim_lists[1][prop]=len(slot)
                else:
                    if len(slot) == 0:
                        continue

                    Elem_and_ElemPrim_lists[0][prop]=len(filter(lambda x: x in self.query_set,slot))
                
                
        return elem_property_lists
    
    def __isPrimitive(self, prop):
        return prop.getTarget().isDatatypeDefinition()
    
    def addIgnored(self, qelem, prop):

        # Isn't there a function for this???
        prop_type=None
        for prop_ed in qelem.__properties__():
            if prop_ed.getName() == prop.name:
                prop_type=prop_ed
                break

        assert prop_type

        property_lists = self.elem_property_lists[qelem]

        for p_list in property_lists:
            if prop_type in p_list:
                p_list.pop(prop_type)
                break

    def setElemCmpFun(self, elem, fun):
        assert elem in self.query_set
        self.elem_cmp_fun[elem]=fun

    def setElemDisjoint(self, elem):
        assert elem in self.query_set
        self.disjoint_elements.append(elem)
    
class QueryState:
    """class QueryState(match_data):
    \tImplements visitor and holds the match data used by search.
    \tInput to match, results are retrieved from the QueryState
    \tinstance

    \tThe visitor function is called for each match and by returning
    \tTrue or False it stops or continues matching. Visitor calls
    \tsaveResult to save the match so that it can be retrieved after
    \tmatching everything."""
    
    def __init__(self, match_data):
        self.q2t={}
        self.stack=[]
        self.results=[]
        self.match_data = match_data
        self.disjoint_match={}
        for disj_elem in match_data.disjoint_elements:
            self.disjoint_match[disj_elem]=[]

    def reset(self):
        self.q2t={}
        self.stack=[]

    def primitive_properties(self, elem):
        return self.match_data.elem_property_lists[elem][1]

    def element_properties(self, elem):
        return self.match_data.elem_property_lists[elem][0]

    def push(self, q_elem, t_elem):
        self.stack.append((q_elem, t_elem))

    def pop(self):
        q_elem,t_elem=self.stack[-1]
        del self.stack[-1]
        return q_elem,t_elem
            
    def stack_size(self):
        return len(self.stack)

    def clone(self):
        state=copy.copy(self)
        state.q2t=self.q2t.copy()
        state.stack=self.stack[:]
        #assert self.match_data.cmp_elem_fun == state.match_data.cmp_elem_fun
        
        return state
            
    def addPair(self, query_item, target_item):
        assert not self.q2t.has_key(query_item), " Cyclic match :-("+unicode(query_item)
        assert not target_item in self.q2t.values()
        assert not isinstance(query_item, coral.ElementPrimitive)
        self.q2t[query_item]=target_item

    def isGoal(self):
        return len(self.q2t) == self.match_data.query_size

    def saveResult(self):
        # q2t Doesn't need copying, since it is copied in clone
        # self.results.append(self.q2t.copy())
        for disj_elem, matches in self.disjoint_match.items():
            if self.q2t[disj_elem] in matches:
                return

        for k,v in self.disjoint_match.items():
            self.disjoint_match[k].append(self.q2t[k])

        self.results.append(self.q2t)

    def getResults(self):
        return self.results

    def isFeasiblePair(self,q_elem, t_elem):
        fun = self.match_data.elem_cmp_fun.get(q_elem, None)
        if fun:
            return fun(q_elem, t_elem)
                
        return self.match_data.cmp_elem_fun(q_elem, t_elem)
    
    def visitor(self):
        self.saveResult()
        return False # False makes the matching routine continue looking for more matches


class OneMatchState(QueryState):
    """Using this State will yield a maximum of one result"""
    def visitor(self):
        self.saveResult()
        return True

class TwoMatchState(QueryState):
    """Using this State will yield a maximum of two results"""
    def visitor(self):
        self.saveResult()
        if len(self.results) > 1:
            return True
        return False


# NEW search planning state, to be used with match_plan from query.py

class StatePlan:
    def __init__(self, match_data):
        self.q2t = {}
        self.stack = []
        self.results = []
        self.match_data = match_data
        self.disjoint_match = {}
        self.matching_negative = False
        for disj_elem in match_data.disjoint_elements:
            self.disjoint_match[disj_elem] = []

        self.searchPlan = SearchPlan(match_data)
        # self.searchPlan = planner.breadthFirstPlan(self.match_data)
        
    def addPair(self, q_elem, t_elem):
        assert not self.q2t.has_key(q_elem), " Cyclic match :-("+unicode(q_elem)
        assert not t_elem in self.q2t.values()
        assert not isinstance(q_elem, coral.ElementPrimitive)
        self.q2t[q_elem]=t_elem

    def isFeasiblePair(self, q_elem, t_elem):
        fun = self.match_data.elem_cmp_fun.get(q_elem, None)
        if fun:
            return fun(q_elem, t_elem)
                
        return self.match_data.cmp_elem_fun(q_elem, t_elem)

    def clone(self):
        state=copy.copy(self)
        state.q2t=self.q2t.copy()
        assert self.match_data.cmp_elem_fun == state.match_data.cmp_elem_fun
        
        return state

    def getResults(self):
        return self.results

    def primitive_properties(self, elem):
        return self.match_data.elem_property_lists[elem][1]

    def element_properties(self, elem):
        return self.match_data.elem_property_lists[elem][0]
    
    def nextQueryNode(self):
        if len(self.q2t) < len(self.searchPlan):
            return self.searchPlan.at(len(self.q2t))
        else:
            return None
        
    def saveResult(self):
        # q2t Doesn't need copying, since it is copied in clone
        # self.results.append(self.q2t.copy())
        for disj_elem, matches in self.disjoint_match.items():
            if self.q2t[disj_elem] in matches:
                return

        for k,v in self.disjoint_match.items():
            self.disjoint_match[k].append(self.q2t[k])

        self.results.append(self.q2t)

    def visitor(self):
        #print "------->visitor<-------"
        self.saveResult()

        return False

    def isGoal(self):
        return len(self.q2t) == self.match_data.query_size
    
    def check_constraints(self, q_elem, t_elem):
        # Primitive
        # Start by matching only the slots which contain ElementPrimitives.
        for primitive_prop in self.primitive_properties(q_elem):
            propName=primitive_prop.getName()
            q_slot=q_elem.__getSlotByName__(propName)
            t_slot=t_elem.__getSlotByName__(propName)
            
            for q_item,t_item in map(None,q_slot,t_slot):
                if not q_item == t_item:
                    #print "Primitive bad"
                    return False

        # Adjacent Element
        # These constraints are known by the search plan and can be made faster
        for element_prop in self.element_properties(q_elem).iterkeys():
            propName = element_prop.getName()
            q_slot = q_elem.__getSlotByName__(propName)
            t_slot = t_elem.__getSlotByName__(propName)
            
            for q_item in q_slot:
                if self.q2t.has_key(q_item):
                    if not self.q2t[q_item] in t_slot:
                        #print "Slot",propName,"not good in", t_elem
                        #print "q_slot",q_slot
                        #print "t_slot",t_slot
                        return False

        return True

    def start_matching_negative(self):
        if self.matching_negative:
            return False

        self.matching_negative = True

        # init one match state

        # Add negative elements to query_set and append negative searchplan
        self.searchPlan.useNegative()
        
        return True
    
    def stop_matching_negative(self):
        self.matching_negative = False
        self.searchPlan.useNegative(False)
